from typing import Mapping
from typing import Any

from matplotlib.axes import Axes
from matplotlib.collections import PolyCollection
from matplotlib.dates import date2num
import numpy as np

from fintime.abc import Artist
from fintime.types import Volume
from fintime.artists.utils import get_rectangle_vertices, to_num_td


class Volume(Artist):
    def __init__(
        self,
        data: Volume = {},
        twinx: bool = False,
        config: Mapping[str, Any] = {},
    ):
        self._data = data
        self._twinx = twinx
        self._cfg = config

    def get_width(self) -> float:
        if self._cfg.volume.panel.width:
            return self._cfg.volume.panel.width
        return self._cfg.volume.panel.width_per_bar * self._data["dt"].size

    def get_height(self) -> float:
        return self._cfg.volume.panel.height

    def get_xmin(self) -> np.datetime64:
        return self._data["dt"][0]

    def get_xmax(self) -> float:
        td = self._data["dt"][-1] - self._data["dt"][-2]
        return self._data["dt"][-1] + td

    def get_ymin(self) -> float:
        return 0.0

    def get_ymax(self) -> float:
        max_vol = max(self._data["vol"])
        return max_vol + max_vol * self._cfg.volume.padding.ymax

    def draw(self, axes: Axes) -> Axes:
        b_xmin = date2num(self._data["dt"])
        b_num_td = to_num_td(self._data["dt"])
        b_vol = self._data["vol"]
        b_close = self._data["close"]
        b_close_last = np.roll(b_close, 1)

        b_xmax = b_xmin + b_num_td * self._cfg.candlestick.body.relwidth
        b_ymin = np.zeros_like(b_vol)
        b_ymax = b_vol

        indices_with_volume = np.where(b_ymax)[0]

        body_verts = get_rectangle_vertices(
            b_xmin=b_xmin[indices_with_volume],
            b_xmax=b_xmax[indices_with_volume],
            b_ymin=b_ymin[indices_with_volume],
            b_ymax=b_ymax[indices_with_volume],
        )

        price_diff = b_close - b_close_last
        c_up = np.where(price_diff > 0, self._cfg.volume.facecolor.up, "")
        c_down = np.where(price_diff < 0, self._cfg.volume.facecolor.down, "")
        c_flat = np.where(price_diff == 0, self._cfg.volume.facecolor.flat, "")
        facecolors = np.char.add(np.char.add(c_up, c_down), c_flat)
        facecolors[0] = self._cfg.volume.facecolor.flat

        bodies = PolyCollection(
            verts=body_verts,
            facecolors=facecolors,
            zorder=self._cfg.candlestick.zorder,
            alpha=self._cfg.candlestick.body.alpha,
        )

        axes.add_artist(bodies)

        return axes



from typing import Mapping
from typing import Any

from matplotlib.axes import Axes
from matplotlib.collections import LineCollection, PolyCollection
from matplotlib.dates import date2num
import numpy as np
from fintime.types import Ohlc

from fintime.artists.utils import (
    get_rectangle_vertices,
    get_horizontal_line_segments,
    get_vertical_line_segments,
    to_num_td,
)
from fintime.abc import Artist


class CandleStick(Artist):
    def __init__(
        self, data: Ohlc = {}, twinx: bool = False, config: Mapping[str, Any] = {}
    ):
        self._data = data
        self._twinx = twinx
        self._cfg = config

    def get_width(self) -> float:
        if self._cfg.candlestick.panel.width:
            return self._cfg.candlestick.panel.width
        return self._cfg.candlestick.panel.width_per_bar * self._data["dt"].size

    def get_height(self) -> float:
        return self._cfg.candlestick.panel.height

    def get_xmin(self) -> np.datetime64:
        return self._data["dt"][0]

    def get_xmax(self) -> np.datetime64:
        td = self._data["dt"][-1] - self._data["dt"][-2]
        return self._data["dt"][-1] + td

    def get_ymin(self) -> float:
        padding = self._cfg.candlestick.padding.ymin
        delta = max(self._data["high"]) - min(self._data["low"])
        return min(self._data["low"]) - delta * padding

    def get_ymax(self) -> float:
        padding = self._cfg.candlestick.padding.ymax
        delta = max(self._data["high"]) - min(self._data["low"])
        ymax = max(self._data["high"]) + delta * padding
        return ymax

    def draw(self, axes: Axes) -> Axes:
        b_xmin = date2num(self._data["dt"])
        b_num_td = to_num_td(self._data["dt"])
        b_open = self._data["open"]
        b_high = self._data["high"]
        b_low = self._data["low"]
        b_close = self._data["close"]

        b_xmax = b_xmin + b_num_td * self._cfg.candlestick.body.relwidth
        b_ymin = np.minimum(b_open, b_close)
        b_ymax = np.maximum(b_open, b_close)

        indices_with_doji = np.where(b_open == b_close)[0]
        indices_without_doji = np.where(b_open != b_close)[0]

        body_verts = get_rectangle_vertices(
            b_xmin=b_xmin[indices_without_doji],
            b_xmax=b_xmax[indices_without_doji],
            b_ymin=b_ymin[indices_without_doji],
            b_ymax=b_ymax[indices_without_doji],
        )

        doji_segs = get_horizontal_line_segments(
            b_xmin=b_xmin[indices_with_doji],
            b_xmax=b_xmax[indices_with_doji],
            b_y=b_ymin[indices_with_doji],
        )

        b_x = (b_xmin + b_xmax) * 0.5
        indices_with_upper_wick = np.where(b_ymin != b_low)[0]
        indices_with_lower_wick = np.where(b_ymax != b_high)[0]

        upper_wick_segs = get_vertical_line_segments(
            b_x=b_x[indices_with_upper_wick],
            b_ymin=b_low[indices_with_upper_wick],
            b_ymax=b_ymin[indices_with_upper_wick],
        )
        lower_wick_segs = get_vertical_line_segments(
            b_x=b_x[indices_with_lower_wick],
            b_ymin=b_ymax[indices_with_lower_wick],
            b_ymax=b_high[indices_with_lower_wick],
        )

        wick_segs = np.concatenate((upper_wick_segs, lower_wick_segs), axis=0)

        facecolors = np.where(
            b_close[indices_without_doji] >= b_open[indices_without_doji],
            self._cfg.candlestick.body.up_color,
            self._cfg.candlestick.body.down_color,
        )

        bodies = PolyCollection(
            verts=body_verts,
            facecolors=facecolors,
            zorder=self._cfg.candlestick.zorder,
            alpha=self._cfg.candlestick.body.alpha,
        )

        wicks = LineCollection(
            segments=wick_segs,
            colors=self._cfg.candlestick.wick.color,
            linewidths=self._cfg.candlestick.wick.linewidth,
            zorder=self._cfg.candlestick.zorder,
        )

        dojis = LineCollection(
            segments=doji_segs,
            colors=self._cfg.candlestick.wick.color,
            linewidths=self._cfg.candlestick.doji.linewidth,
            zorder=self._cfg.candlestick.zorder,
        )
        axes.add_artist(bodies)
        axes.add_artist(wicks)
        axes.add_artist(dojis)



import numpy as np
from matplotlib.dates import (
    DateFormatter,
    DayLocator,
    HourLocator,
    MinuteLocator,
    MonthLocator,
    SecondLocator,
    YearLocator,
)
from fintime.types import Timezone


Locator = DayLocator | HourLocator | YearLocator | MonthLocator | YearLocator


def get_nearest(arr: np.ndarray, value):
    return arr.flat[np.abs(arr - value).argmin()]


def to_secs(td: np.timedelta64) -> int:
    return td.astype("timedelta64[s]").astype(np.int32)


def get_secs_per_inch(ax_span, ax_rel_width: float, width: float):
    ax_width = ax_rel_width * width
    return (ax_span / ax_width).astype(np.int32)


def make_get_locator_and_formatter(tz: Timezone):
    tds_args = [
        (1, "s"),
        (2, "s"),
        (5, "s"),
        (10, "s"),
        (15, "s"),
        (20, "s"),
        (30, "s"),
        (1, "m"),
        (2, "m"),
        (5, "m"),
        (10, "m"),
        (15, "m"),
        (20, "m"),
        (30, "m"),
        (1, "h"),
        (2, "h"),
        (4, "h"),
        (1, "D"),
        (1, "W"),
        (1, "M"),
        (3, "M"),
        (1, "Y"),
    ]

    bins = {}
    secs = []

    for td_args in tds_args:
        td = np.timedelta64(*td_args)
        n_secs = to_secs(td=td)
        if n_secs < to_secs(np.timedelta64(1, "m")):
            locator = SecondLocator(range(0, 60, n_secs), tz=tz)
            formatter = DateFormatter("%H:%M:%S", tz=tz)

        elif n_secs < to_secs(np.timedelta64(60, "m")):
            step = int(n_secs / 60)
            locator = MinuteLocator(range(0, 60, step), tz=tz)
            formatter = DateFormatter("%H:%M", tz=tz)

        elif n_secs < to_secs(np.timedelta64(4, "h")):
            step = int(n_secs / 3600)
            locator = HourLocator(range(0, 24, step), tz=tz)
            formatter = DateFormatter("%d-%H", tz=tz)

        elif n_secs < to_secs(np.timedelta64(1, "D")):
            locator = DayLocator(tz=tz)
            formatter = DateFormatter("%m-%d", tz=tz)
        else:
            locator = DayLocator(interval=5, tz=tz)
            formatter = DateFormatter("%m-%d", tz=tz)

        bins[n_secs] = {"locator": locator, "formatter": formatter}
        secs.append(n_secs)
    secs = np.array(secs)

    def get_locator_and_formatter(
        secs_per_inch: float, nudge: int = 0
    ) -> tuple[Locator, DateFormatter]:
        td_secs = get_nearest(arr=secs, value=secs_per_inch)

        # increase spacing for tick labels that contain more characters
        if td_secs <= 60:
            nudge += 1

        if nudge:
            i = np.where(secs == td_secs)[0][0]
            if nudge > 0:
                i = min(secs.size - 1, i + nudge)
            else:
                i = max(0, i + nudge)
            td_secs = secs[i]

        return bins[td_secs]["locator"], bins[td_secs]["formatter"]

    return get_locator_and_formatter


def get_major_locator_and_formatter(
    x_min: np.datetime64,
    x_max: np.datetime64,
    width: float,
    tz: Timezone,
    nudge: int = 0,
):
    get_locator_and_formatter = make_get_locator_and_formatter(tz=tz)
    secs_per_inch = (x_max - x_min).astype("timedelta64[s]").astype(np.int64) / width
    return get_locator_and_formatter(secs_per_inch=secs_per_inch, nudge=nudge)



from typing import Optional, Callable, Any, Mapping
from abc import ABC, abstractmethod
from copy import deepcopy

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.gridspec import SubplotSpec
from fieldconfig import Config


class XLim(ABC):
    """
    Specifies the shared x-axis for Artists, Panels, and Subplots.
    """

    @abstractmethod
    def get_xmin(self) -> float: ...

    @abstractmethod
    def get_xmax(self) -> float: ...


class YLim(ABC):
    """
    Specifies the shared y-axis for Artists and Panels.
    """

    @abstractmethod
    def get_ymin(self, twinx: bool = False) -> float: ...

    @abstractmethod
    def get_ymax(self, twinx: bool = False) -> float: ...


class Sized(ABC):
    """
    Specifies the shared protocol for Artists, Panels, Subplots, and Grid.
    """

    @abstractmethod
    def get_width(self) -> float:
        """Return width in inches."""

    @abstractmethod
    def get_height(self) -> float:
        """Return height in inches."""

    @abstractmethod
    def draw(self, canvas: Figure | SubplotSpec | Axes) -> None:
        """Draw the component on the specified canvas."""

    def _set_size(self, width: float, height: float) -> None:
        """Set the components width and height in inches"""

        self._width = width
        self._height = height

    def _update_data(self, data: Mapping[str, Any]):
        """Update the inherited data with local updates, if any are present."""

        data.update(self._data)
        self._data = data

    def _update_config(self, config: Config):
        """Update the inherited Config with local updates, if any are present."""

        config.update(self._cfg)
        self._cfg = config


class Artist(Sized, XLim, YLim):
    def is_twinx(self):
        return self._twinx

    def propagate_data(self, data: Mapping[str, Any]) -> None:
        self._update_data(data=data)

    def propagate_config(self, config: Config) -> None:
        self._update_config(config=config)

    def propagate_size(self, width: float, height: float) -> None:
        self._set_size(width=width, height=height)


class Composite(Sized):
    def __init__(
        self,
        components: list[Sized],
        width: Optional[float],
        height: Optional[float],
        op_heights: Callable[[float], float],
        op_widths: Callable[[float], float],
    ):
        self._components = components
        self._width = width
        self._height = height
        self._op_widths = op_widths
        self._op_heights = op_heights

    @property
    def nrows(self):
        return len(self._components)

    def _get_heights(self) -> list[float]:
        return [comp.get_height() for comp in self._components]

    def _get_widths(self) -> list[float]:
        return [comp.get_width() for comp in self._components]

    def get_height(self) -> float:
        if self._height:
            return self._height
        return self._op_heights(self._get_heights())

    def get_width(self) -> float:
        if self._width:
            return self._width
        return self._op_widths(self._get_widths())

    def propagate_data(self, data: Mapping[str, Any]) -> None:
        self._update_data(data=data)

        for comp in self._components:
            comp.propagate_data(data=deepcopy(data))

    def propagate_config(self, config: Config) -> None:
        self._update_config(config=config)

        for comp in self._components:
            comp.propagate_config(config=config.copy())

    def propagate_size(self, width: float, height: float) -> None:
        self._set_size(width=width, height=height)

        old_height = self.get_height()
        scalar = height / old_height
        new_heights = [h * scalar for h in self._get_heights()]

        for comp, new_height in zip(self._components, new_heights):
            comp.propagate_size(width, new_height)



import numpy as np


def generate_random_trade_ticks(
    dt_start: np.datetime64 = np.datetime64("now"),
    td_session: np.timedelta64 = np.timedelta64(6, "h"),
    initial_price=100,
    volatility: float = 0.5,
    liquidity: float = 2.5,
    seed=None,
):

    assert td_session < np.timedelta64(4, "D")
    assert volatility > 0 and volatility <= 10
    assert liquidity > 0 and liquidity <= 5

    dt_stop = dt_start + td_session
    dt_range = np.arange(dt_start, dt_stop, 10, dtype="datetime64[ms]")
    n_secs_session = td_session / np.timedelta64(1, "s")
    np.random.seed(seed)
    n_ticks = int(n_secs_session * liquidity**2)

    dt_ticks = np.random.choice(dt_range, size=n_ticks, replace=False)
    dt_ticks = np.sort(dt_ticks)
    scale = 0.001 * volatility**2
    drift_values = np.random.normal(loc=0, scale=scale, size=n_ticks)

    cumulative_drift = np.cumsum(drift_values)
    prices = initial_price * np.exp(cumulative_drift)
    sizes = np.random.randint(0, 100, size=n_ticks)
    ticks = {"dt": dt_ticks, "price": prices, "size": sizes}
    return ticks


def to_timebar(ticks, span: int, n_bars: int = 200):
    t_dt = ticks["dt"]
    t_price = ticks["price"]
    t_size = ticks["size"]

    td_step = np.timedelta64(span, "s")
    dt_start = t_dt[0].astype(f"datetime64[{span}s]")
    dt_stop = t_dt[-1].astype(f"datetime64[{span}s]") + td_step

    b_dt = np.arange(dt_start, dt_stop, td_step)
    b_open = []
    b_high = []
    b_low = []
    b_close = []
    b_vol = []

    i_tick = 0
    p_close = t_price[0]
    for dt_open in b_dt:
        prices = []
        volume = 0
        while t_dt[i_tick] < dt_open:
            prices.append(t_price[i_tick])
            volume += t_size[i_tick]
            i_tick += 1
        i_tick += 1

        if prices:
            b_open.append(prices[0])
            b_high.append(max(prices))
            b_low.append(min(prices))
            b_close.append(prices[-1])
            b_vol.append(volume)
            p_close = prices[-1]
        else:
            b_open.append(p_close)
            b_high.append(p_close)
            b_low.append(p_close)
            b_close.append(p_close)
            b_vol.append(0)

    return {
        "dt": np.array(b_dt[-n_bars:]).astype("datetime64[ms]"),
        "open": np.array(b_open[-n_bars:]).round(2),
        "high": np.array(b_high[-n_bars:]).round(2),
        "low": np.array(b_low[-n_bars:]).round(2),
        "close": np.array(b_close[-n_bars:]).round(2),
        "vol": np.array(b_vol[-n_bars:]),
    }


# tt = generate_random_ticks()
# y = to_timebar(1, trade_ticks=tt)
# # %%
# def gen_tick_data(
#     start: np.datetime64,
#     td: np.timedelta64,
#     initial_price=100,
#     vollatilty=8,
#     liquidity=10,
#     seed=None,
# ):
#     np.random.seed(seed)
#     n_ticks = int(np.random.random() * 10000)
#     dt_arange = np.arange(start, start + td, dtype="datetime64[ms]")
#     dts = np.random.choice(dt_arange, size=n_ticks, replace=False)
#     dts = np.sort(dts)
#     drift_values = np.random.normal(loc=0, scale=0.001, size=n_ticks)
#     cumulative_drift = np.cumsum(drift_values)
#     prices = initial_price * np.exp(cumulative_drift)
#     sizes = np.random.randint(0, 100, size=prices.size) * 1000
#     ticks = {"dt": dts, "price": prices, "size": sizes}

#     return ticks


# tick_data = gen_tick_data(np.datetime64("now"), np.timedelta64(1, "h"))


# def to_timebar(ticks, span: int): ...



from copy import copy
from typing import Any, Mapping

import numpy as np

from matplotlib.dates import date2num
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec, SubplotSpec, GridSpecFromSubplotSpec
from matplotlib.pyplot import Axes

from fintime.abc import Composite, Artist, XLim, YLim
from fintime.config import Config
from fintime.ticks import get_major_locator_and_formatter
from fintime.types import SizeSpec


class Panel(Composite, XLim, YLim):
    def __init__(
        self,
        artists: list[Artist],
        data: dict = {},
        size: SizeSpec = (None, None),
        config: Mapping[str, Any] = {},
    ):
        super().__init__(
            components=artists,
            width=size[0],
            height=size[1],
            op_widths=max,
            op_heights=sum,
        )
        self._artists = artists
        self._cfg = Config(config)
        self._data = data

    def draw(self, axes: Axes):
        axes.set_ylim(self.get_ymin(), self.get_ymax())
        axes.set_facecolor(self._cfg.panel.facecolor)

        for artist in self._artists:
            artist.draw(axes)

    def get_xmin(self) -> float:
        return min([comp.get_xmin() for comp in self._components])

    def get_xmax(self) -> float:
        return min([comp.get_xmax() for comp in self._components])

    def get_ymin(self, twinx: bool = False) -> float:
        return min(
            [comp.get_ymin() for comp in self._components if comp.is_twinx() == twinx]
        )

    def get_ymax(self, twinx: bool = False) -> float:
        return min(
            [comp.get_ymax() for comp in self._components if comp.is_twinx() == twinx]
        )


class Subplot(Composite, XLim):
    def __init__(
        self,
        panels: list[Panel],
        data: dict = {},
        size: SizeSpec = (None, None),
        config: Mapping[str, Any] = {},
    ):
        super().__init__(
            components=panels,
            width=size[0],
            height=size[1],
            op_widths=max,
            op_heights=sum,
        )
        self._cfg = Config(config)
        self._data = data

    def get_xmin(self) -> float:
        return min([comp.get_xmin() for comp in self._components])

    def get_xmax(self) -> float:
        return min([comp.get_xmax() for comp in self._components])

    def get_ymin(self) -> float:
        return min([comp.get_ymin() for comp in self._components])

    def get_ymax(self) -> float:
        return max([comp.get_ymax() for comp in self._components])

    def draw(self, canvas: Figure | SubplotSpec):

        if isinstance(canvas, SubplotSpec):

            gs = GridSpecFromSubplotSpec(
                nrows=self.nrows,
                ncols=1,
                subplot_spec=canvas,
                height_ratios=self._get_heights(),
                hspace=0.0,
                wspace=0,
            )
        else:
            gs = GridSpec(
                nrows=self.nrows,
                ncols=1,
                height_ratios=self._get_heights(),
                figure=canvas,
                hspace=0,
                wspace=0.01,
            )

        xmin = self.get_xmin()
        xmax = self.get_xmax()

        locator, formatter = get_major_locator_and_formatter(
            x_min=xmin,
            x_max=xmax,
            width=self.get_width(),
            nudge=self._cfg.xaxis.tick.nudge,
            tz=self._cfg.timezone,
        )
        if isinstance(xmin, np.datetime64):
            xmin = date2num(xmin)
            xmax = date2num(xmax)

        nrows, ncols = gs.get_geometry()
        axes_list = gs.subplots()
        if isinstance(axes_list, Axes):
            axes_list = [axes_list]

        for panel, axes in zip(self._components, axes_list):

            rowspan = axes.get_subplotspec().rowspan
            labeltop = rowspan.start == 0
            labelbottom = rowspan.stop == nrows
            axes.set_xlim(xmin, xmax)
            axes.grid(linestyle="-", alpha=0.5, zorder=10)
            axes.tick_params(
                which="both", labelsize=12, top=labeltop, bottom=labelbottom, right=True
            )
            axes.yaxis.set_tick_params(right=True, labelright=True)
            axes.xaxis.set_tick_params(
                labeltop=labeltop, labelbottom=labelbottom, direction="out"
            )
            axes.xaxis.set_major_locator(locator=locator)
            axes.xaxis.set_major_formatter(formatter=formatter)
            panel.draw(axes)

        # enforce x-axis panel alignment when interatively viewed.
        for i in range(1, len(axes_list)):
            axes_list[i].sharex(axes_list[i - 1])


class Grid(Composite):
    def __init__(
        self,
        subplots: list[Subplot],
        size: SizeSpec,
        data,
        config,
    ):
        super().__init__(
            components=subplots,
            width=size[0],
            height=size[1],
            op_widths=max,
            op_heights=sum,
        )

        self._data = data
        self._cfg = config

    def draw(self, canvas):
        # TODO: improve dynamic hspace logic
        gridspec = GridSpec(
            nrows=self.nrows,
            ncols=1,
            height_ratios=self._get_heights(),
            figure=canvas,
            hspace=4 / self._height + 0.05,
            wspace=0,
        )
        for subplot, subplot_spec in zip(self._components, gridspec):
            subplot.draw(canvas=subplot_spec)



from typing import TypedDict
from typing import Mapping
from typing import Optional
from typing import Any
from typing import Literal
from fieldconfig import Config

from nptyping import NDArray, Shape
from nptyping import Floating, Datetime64, Integer

Array1D = NDArray[Shape["*"], Any]
FloatArray1D = NDArray[Shape["*"], Floating]
IntArray1D = NDArray[Shape["*"], Integer]
DatetimeArray1D = NDArray[Shape["*"], Datetime64]

SizeSpec = tuple[Optional[float], Optional[float]]

from typing import TypeVar
from datetime import timezone
from dateutil import tz

Timezone = TypeVar("Timezone", timezone, tz.tz.tzfile)

Side = Literal["buy", "sell"]


class Ohlc(TypedDict):
    dt: DatetimeArray1D  # represents the time of the open
    open: FloatArray1D
    high: FloatArray1D
    low: FloatArray1D
    close: FloatArray1D


class Volume(TypedDict):
    dt: DatetimeArray1D  # represents the time of the open
    vol: IntArray1D


class Fill(TypedDict):
    dt: Datetime64
    side: Side
    price: float
    size: int



import numpy as np
from fintime.types import FloatArray1D


def get_vertical_line_segments(
    b_x: FloatArray1D, b_ymin: FloatArray1D, b_ymax: FloatArray1D
):
    return np.stack(
        (
            np.stack((b_x, b_ymin), axis=1),
            np.stack((b_x, b_ymax), axis=1),
        ),
        axis=1,
    )


def get_horizontal_line_segments(
    b_xmin: FloatArray1D, b_xmax: FloatArray1D, b_y: FloatArray1D
):
    return np.stack(
        (
            np.stack((b_xmin, b_y), axis=1),
            np.stack((b_xmax, b_y), axis=1),
        ),
        axis=1,
    )


def get_rectangle_vertices(
    b_xmin: FloatArray1D,
    b_xmax: FloatArray1D,
    b_ymin: FloatArray1D,
    b_ymax: FloatArray1D,
):
    return np.stack(
        (
            np.stack((b_xmin, b_ymin), axis=1),
            np.stack((b_xmin, b_ymax), axis=1),
            np.stack((b_xmax, b_ymax), axis=1),
            np.stack((b_xmax, b_ymin), axis=1),
        ),
        axis=1,
    )


def to_num_td(dt) -> np.timedelta64:
    "infer bar duration in days"
    td = dt - np.roll(dt, 1)
    td[0] = np.median(td)
    return td / np.timedelta64(1, "D")



import warnings
from typing import Any, Mapping, Optional

import matplotlib.pyplot as plt

from fintime.config import get_config
from fintime.types import Array1D, SizeSpec
from fintime.composites import Subplot, Panel, Grid


def plot(
    specs: list[Subplot] | list[Panel],
    data: Mapping[str, Array1D] = {},
    config: Mapping[str, Any] = None,
    figsize: SizeSpec = (None, None),
    title: Optional[str] = None,
    save: Optional["str"] = None,
    rtn: bool = True,
):
    """
    Create a plot using specified subplots or panels.

    Parameters:
    - specs (list[Subplot] | list[Panel]): List of Subplot or Panel instances
    defining the layout of the plot.
    - data (Mapping[str, Array1D], optional): Data to be plotted. Default is an
    empty dictionary.
    - config (Mapping[str, Any], optional): Configuration settings for the plot.
    Default is None.
    - figsize (SizeSpec, optional): Size of the figure. Default is (None, None),
    (size is inferred from subcomponents in specs)
    - title (str, optional): Title for the entire plot. Default is None (no title).
    - save (str, optional): File path to save the plot. Default is None (not saved).
    - rtn (bool, optional): Whether to return the Figure instance. Default is True.

    Returns:
    - Optional[Figure]: The Figure instance if rtn is True, else None.

    Note:
    - Configurations and data are propagated downward to subcomponents, allowing
    for overrides at any level.
    """

    if config is None:
        config = get_config()

    class_name = Subplot if isinstance(specs[0], Panel) else Grid
    gridspec = class_name(specs, size=figsize, data=data, config=config)

    gridspec.propagate_config(config=config)
    gridspec.propagate_data(data=data)

    fig_width = gridspec.get_width()
    fig_height = gridspec.get_height()

    print(fig_width, fig_height)

    gridspec.propagate_size(width=fig_width, height=fig_height)

    fig = plt.figure(
        figsize=(fig_width, fig_height),
        layout=config.figure.layout,
        facecolor=config.figure.facecolor,
    )
    if title:
        fig.suptitle(
            title,
            fontsize=config.figure.title.fontsize,
            fontweight=config.figure.title.fontweight,
            y=config.figure.title.y,
        )
    gridspec.draw(canvas=fig)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="This figure includes Axes that are not compatible with tight_layout",
        )
        if save:
            fig.savefig(save)
    if rtn:
        return fig



from typing import Any
from fieldconfig import Config, Field
from dateutil import tz
import matplotlib.colors as mcolors


def is_positive_number(x):
    return x > 0


def is_between_zero_and_one(x):
    return x >= 0 and x <= 1


def is_valid_mpl_color(x):
    try:
        mcolors.to_rgba(x)
        return True
    except ValueError:
        return False


def get_config():

    cfg = Config(create_intermediate_attributes=True)

    cfg.timezone = tz.gettz("America/New_York")
    cfg.figure.layout = "tight"
    cfg.figure.facecolor = "#f9f9f9"

    cfg.figure.title.fontsize = 22
    cfg.figure.title.fontweight = "bold"
    cfg.figure.title.y = Field(0.98, validator=is_between_zero_and_one)

    cfg.panel.facecolor = "white"
    cfg.xaxis.tick.nudge = 0

    cfg.candlestick.panel.height = Field(9.0, validator=is_positive_number)
    cfg.candlestick.panel.width = Field(None, float)
    cfg.candlestick.panel.width_per_bar = 0.1
    cfg.candlestick.padding.ymin = 0.06
    cfg.candlestick.padding.ymax = 0.06
    cfg.candlestick.zorder = 14
    cfg.candlestick.body.relwidth = Field(0.8, validator=is_between_zero_and_one)
    cfg.candlestick.body.alpha = Field(1.0, validator=is_between_zero_and_one)
    cfg.candlestick.body.up_color = Field("#4EA59A", validator=is_valid_mpl_color)
    cfg.candlestick.body.down_color = Field("#E05D57", validator=is_valid_mpl_color)
    cfg.candlestick.wick.color = Field("#000000", validator=is_valid_mpl_color)
    cfg.candlestick.wick.linewidth = 1.0
    cfg.candlestick.wick.alpha = Field(1.0, validator=is_between_zero_and_one)
    cfg.candlestick.doji.color = Field("#000000", validator=is_valid_mpl_color)
    cfg.candlestick.doji.linewidth = 1.0
    cfg.candlestick.doji.alpha = Field(1.0, validator=is_between_zero_and_one)

    cfg.volume.panel.height = Field(3.0, validator=is_positive_number)
    cfg.volume.panel.width = Field(None, ftype=float)
    cfg.volume.padding.ymax = 0.05
    cfg.volume.facecolor.up = Field("#4EA59A", validator=is_valid_mpl_color)
    cfg.volume.facecolor.down = Field("#E05D57", validator=is_valid_mpl_color)
    cfg.volume.facecolor.flat = Field("#b0b0b0", validator=is_valid_mpl_color)
    cfg.volume.panel.width_per_bar = 0.1

    cfg.disable_intermediate_attribute_creation()
    return cfg
